# BOTJE - Framework to program bots !!
#
# bot/__init__.py (bot package)
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 22-12-2017 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" bot module containing the basic bot classes. """

__version__ = 51
__txt__ = "Framework to program bots !!"

from .dbz import users
from .obj import cdir, cfg, Config, Object
from .evt import Event
from .ldr import Loader
from .log import level
from .shl import parse_cli
from .trc import get_exception

import importlib
import logging
import os
import queue
import select
import sys
import time
import types

READ_ONLY = select.POLLIN | select.POLLPRI | select.POLLHUP | select.POLLERR
READ_WRITE = READ_ONLY | select.POLLOUT
EDGE = select.EPOLLIN  | select.EPOLLOUT | select.EPOLLET

class ENOTIMPLEMENTED(Exception):

    """ the method is not implemented in the inheriting class. """

    pass

class ECHAR(Exception):

    """ characters are not allowed. """
    
    pass
    
_booted = False

nr = 0

class Bot(Loader):

    """ virtual bot class. """

    cc = ""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._nrconnect = 0
        self._outqueue = queue.Queue()
        self._resume = Object()
        self._stopped = False
        self.channels = []
        self.cfg = Config()
        self.id = repr(self)
 
    def announce(self, txt):
        """ print text on joined channels. """
        if cfg.silent:
            return
        if not self.channels:
            self.raw(txt)
        for channel in self.channels:
            self.say(channel, txt)

    def connect(self, user="", passwd=""):
        """ connect to server with user/passwd credentials. """
        return False

    def disconnect(self, e):
        """ disconnect from the server. """
        pass

    def disconnected(self):
        """ disconnected callback. """
        pass

    def event(self, origin="", txt=""):
        e = Event()
        e.id = self.id
        e.origin = origin or self.cfg.server
        e.txt = txt
        return e

    def join(self, channel, password=""):
        """ join a channel. """
        pass

    def joinall(self):
        """ join all channels. """
        for channel in self.channels:
            self.join(channel)

    def nick(self, nick):
        """ set bot's nick. """
        pass

    def out(self, channel, line):
        """ output text on channel. """
        self.say(channel, line)

    def prompt(self, *args, **kwargs):
        """ echo prompt to sys.stdout. """
        pass

    def raw(self, txt):
        """ send txt to server. """
        if cfg.verbose:
            print(txt)
        
    def reconnect(self, event=None):
        """ run a connect loop until connected. """
        while not self._stopped:
            self._nrconnect += 1
            try:
                self.connect()
                break
            except:
                logging.error(get_exception())
            time.sleep(self._nrconnect * 5.0)
            if self._nrconnect > 3:
                 break
        logging.warn("connected to %s" % self.cfg.server)
                         
    def say(self, channel, txt, *args):
        """ say something on a channel. """
        if type(txt) in [list, tuple]:
            txt = ",".join(txt)
        self.raw(txt, *args)

    def start(self, *args, **kwargs):
        """ virtual start method """
        super().start(*args, **kwargs)
        if self.cfg.channel not in self.channels:
            self.channels.append(self.cfg.channel)
        if self not in fleet:
            fleet.append(self)

def boot():
    from bot.ldr import loader
    global cfg
    cfg.update(parse_cli())
    if cfg.banner:
        print("%s #%s - %s\n" % (cfg.name or "BOTJE", __version__, __txt__))
    cdir(cfg.workdir)
    cdir(cfg.logdir)
    if cfg.license:
        m = loader.load_mod("bot.lcs")
        print(m.__license__)
    level(cfg.level, cfg.logdir)
    logging.warn("workdir %s" % cfg.workdir)
    users.meet("root@shell")
    return cfg
    
def get(type="<", name="",  rval=[], g=globals()):
    o = Object()
    for k, v in g.items():
        t = str(v)
        if t.startswith(type):
            if name and name in t:
                o[k] = v
            else:
                o[k] = v
    return o

def root():
    """ return the root directory. """
    path = cfg.workdir
    path = os.path.abspath(path)
    check_permissions(path)
    return path

# VARS

fleet = []
