# BOTJE - Framework to program bots !!
#
# bot/cli.py (command line interface bot)
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 22-12-2017 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" bot module containing the basic bot classes. """

__version__ = 51
__txt__ = "Framework to program bots !!"

from . import fleet
from .dbz import users
from .evt import Event
from .obj import cdir, cfg, Config
from .slt import SelectBot

import logging
import sys

class CLI(SelectBot):

    """ Command Line Interface Bot. """

    cc = ""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._started = False
        self._threaded = True
        self.cfg = Config()
        self.prompted = False
        self.register_fd(sys.stdin)

    def cmnd(self, txt):
        """ execute a command based on txt.  """
        event = Event()
        event.verbose = True
        event.id = self.id
        event.origin = "root@shell"
        event.txt = txt
        event.parse()
        return event

    def event(self):
        """ must return an Event and block for input function. """ 
        e = Event()
        e.id = self.id
        e.cc = self.cc
        e.origin = "root@shell"
        e.txt = input()
        self.prompted = False
        e.txt = e.txt.rstrip()
        return e

    def dispatch(self, event):
        super().dispatch(event)
        event.wait()
        event.prompt()
        
    def prompt(self, *args, **kwargs):
        """ echo prompt to sys.stdout. """
        if self.prompted:
            return
        self.prompted = True
        sys.stdout.write("> ")
        sys.stdout.flush()

    def raw(self, txt):
        """ output txt to sys.stdout """
        try:
            sys.stdout.write(txt)
        except TypeError:
            sys.stdout.write(str(txt))
        sys.stdout.write("\n")
        sys.stdout.flush()

    def start(self, *args, **kwargs):
        super().start(*args, **kwargs)
        self.cfg.fromdisk("cli")
        self.load_mod("bot.cmd")
        self.load_mod("bot.eml")
        if cfg.args:
            event = self.cmnd(" ".join(cfg.args))
            super().dispatch(event)
            self.ready()
            return
        else:
            if cfg.shell:
                self.load_mod("bot.cmd")
                self.prompt()
                self.wait()
        self.ready()

cli = CLI()
            