# BOTJE - Framework to program bots !!
#
# cmd.py (commands)
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 22-12-2017 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" basic commands. """

from . import Bot, get, fleet, __version__, __txt__
from .clk import Timed
from .dbz import db, users
from .evt import Event
from .hdl import HandlerBot
from .ldr import loader
from .obj import cfg as _cfg
from .obj import cache, Config, Object, name, slice, allowchar, allowedchars
from .shl import reset
from .tmd import day, elapsed, to_time, to_day, now, parse_time
from .thr import Launcher, launcher

import _thread
import logging
import json
import mailbox
import string
import sys
import random
import time
import os

psformat = "%-4s %-8s %-60s"

_begintime = time.time()
_starttime = time.time()

def shutdown(event):
    """ close commands plugins. """
    pass

class Entry(Object):
    """ data entry class . """
    pass

class Log(Entry):
    """ a log entry to log what is happening on the day. """
    pass


class Rss(Entry):
    """ rss entry """
    pass

class Shop(Entry):
    """ shopping item for the wife to shop. """
    pass

class Todo(Entry):
    """ todo entry """
    pass

class Tomorrow(Entry):
    """ todo for tomorrow to be done. """
    pass

class Watch(Entry):
    """ files to watch. """
    pass

def randomname():
    s = ""
    for x in range(8):
        s += random.choice(allowedchars)
    return s

def all(event):
    from .ldr import cmds
    events = []
    bot = HandlerBot()
    bot.start()
    bot.load_mod("bot.cmd")
    for cmnd in random.shuffle(bot._cmds):
        if cmnd in ["all", "tests", "tinder"]:
            continue
        e = Event()
        e.origin = event.origin
        e.txt = cmnd + " " + randomname()
        e.txt = e.txt.strip()
        e.cc = ""
        bot.prep(e)
        bot.put(e)
        events.append(e)
    event.reply("%s events pushed" % len(events))
    for e in events:
        e.wait()
    bot.stop()

all.threaded = True
all.perm = "OPER"

def announce(event):
    """ announce text on all channels in fleet. """
    event.announce(event._parsed.rest)
    #event._result.append(event._parsed.rest)

announce.perm = "ANNOUNCE"

def attr(event):
    if not event._parsed.args:
         options = ",".join(os.listdir(_cfg.workdir))
         event.reply("attr <prefix>, use one of %s" % options)
         return
    prefix = event._parsed.args[0]
    try:
        keys = event._parsed.args[1:]
    except:
        keys = []
    obj = db.last(prefix)
    if obj:
        for key in keys:
            obj = getattr(obj, key, None)
        try:
            res = obj.keys()
        except AttributeError:
            res = []
            for o in obj:
                res.extend(o.keys())
            res = set(res)
        event.reply("\n".join(sorted([x for x in res if not x.startswith("_")])))
    else:
        event.reply("no %s prefix found." % event._parsed.args[0])

attr.perm = "USER"

def begin(event):
    """ begin stopwatch. """
    global _begintime
    t = to_time(now())
    _begintime = t
    event.reply("time is %s" % time.ctime(t))

begin.perm = "USER"

def cfg(event):
    """ edit config files. """
    args = event._parsed.args
    if not args:
        event.reply(_cfg)
        return
    name = args[0]
    p = os.path.abspath(os.path.join(_cfg.workdir, "config", name))
    if p in cache:
        obj = cache[p]
    else:
        obj = Config()
        obj.fromdisk(name)
    try:
        key = args[1]
    except IndexError:
        event.reply(obj)
        return
    if key not in obj:
        event.reply("EKEY %s" % key)
        return
    if len(args) == 2:
        event.reply('%s="%s"' % (key, obj[key]))
        return
    val = ""
    if len(args) > 3:
        val = args[2:]
    if len(args) == 3:
        val = args[2]
    try:
        val = int(val)
    except:
        try:
            val = float(val)
        except:
            pass
    if type(val) == list:
        obj[key] = list(val)
    elif val in ["True", "true"]:
        obj[key] = True
    elif val in ["False", "false"]:
        obj[key] = False
    else:
        try:
            val = '"%s"' % val
            obj[key] = json.loads(val)
        except (SyntaxError, ValueError) as ex:
            event.reply("%s: %s" % (str(ex), val))
            return
    if key == "user" and "@" in obj[key]:
        obj["username"], obj["server"] = obj[key].split("@")
    obj.sync(os.path.join(_cfg.workdir, "config", name))
    event.ok(val)

cfg.perm = "CFG"

def cmds(event):
    """ show list of commands. """
    cmnds = get("<func", g=globals())
    event.reply(",".join(sorted(cmnds)))

cmds.perm = "USER"

def deleted(event):
    """ show deleted records. """
    event.nodel = True
    nr = 0
    for obj in db.selected(event):
        if "deleted" in dir(obj) and obj.deleted:
            nr += 1
            event.display(obj, str(nr))

deleted.perm = "USER"

def delperm(event):
    """ delete permissions of an user. """
    try:
        origin, perm = event._parsed.args
    except:
        event.reply("delperm <origin> <perm>")
        return
    result = users.delete(origin, perm)
    if not result:
        event.reply("can't find user %s" % origin)
        return
    event.ok(origin, perm)

delperm.perm = "OPER"

def edit(event):
    """ edit and save objects. """
    mods = get("<module", "bot", g=globals())
    args = event._parsed.args
    if not args:
        event.reply(",".join(mods))
        return
    name = args[0]
    if name not in mods:
        event.reply("%s is not an object in bot package" % name)
        return
    obj = getattr(mods, name)
    try:
        key = args[1]
    except IndexError:
        event.reply(obj)
        return
    if len(args) > 3:
        val = args[2:]
    elif len(args) == 3:
        val = args[2]
    else:
        event.reply("%s: %s" % (key, value))
        return
    event.ok(key, val)

edit.perm = "EDIT"

def end(event):
    """ stop stopwatch. """
    diff = time.time() - _begintime
    if diff:
        event.reply("time elapsed is %s" % elapsed(diff))

end.perm = "USER"

def exit(event):
    """ stop the bot. """
    _thread.interrupt_main()

exit.perm = "EXIT"

def fix(event):
    """ fix a object by loading and saving it. """
    fn = event._parsed.rest
    fn = os.path.abspath(fn)
    if not fn:
        event.reply("fix <path>")
        return
    if not os.path.isfile(fn):
        event.reply("%s is not a file" % fn)
        return
    o = Object()
    o.load(fn)
    p = o.save()
    event.reply("saved %s" % p)

fix.perm = "OPER"

def find(event):
    """ present a list of objects based on prompt input. """
    nr = 0
    seen = []
    for obj in db.selected(event):
        for key in event._parsed.uniqs.keys():
            val =  getattr(obj, key, None)
            if val and val in seen:
                continue
            else:
                seen.append(val)
        if "d" in event._parsed.enabled:
            event.reply(obj)
        else:
            event.display(obj, str(nr))
        nr += 1

find.perm = "USER"

def first(event):
    """ show the first record matching the given criteria. """
    obj = db.first(*event._parsed.args)
    if obj:
        event.reply(obj)

first.perm = "USER"

def last(event):
    """ show last objectect matching the criteria. """
    if not event._parsed.args:
        event.reply("last <prefix> <value>")
        return
    obj = db.last(*event._parsed.args)
    if obj:
        event.reply(obj)

last.perm = "USER"

def load(event):
    """ force a plugin reload. """
    bot = fleet[0]
    if not event._parsed.rest:
        event.reply(",".join([x.split(".")[-1] for x in bot.modules("bot")]))
        return
    for modname in bot.modules("bot"):
        if event._parsed.rest not in modname:
            continue
        try:
            mod = bot.reload(modname, True)
            event.reply("load %s" % modname)
        except (AttributeError, KeyError) as ex:
            event.reply("%s %s" % (modname, str(ex)))

load.perm = "LOAD"

def log(event):
    """ log some text. """
    if not event._parsed.rest:
        event.reply("log <item>")
        return
    o = Log()
    o.log = event._parsed.rest
    o.save()
    event.ok(1)

log.perm = "USER"

def loud(event):
    """ disable silent mode of a bot. """
    for bot in fleet:
        if event.id == bot.id:
            bot.cfg.silent = False
            bot.cfg.sync()
            event.reply("silent mode disabled.")

loud.perm = "LOUD"

def loop(event):
    try:
        nr = int(event._parsed.args[0])
    except:
        nr = 10
    for x in range(nr):
        launcher.launch(all, event)

loop.perm = "TEST"

def ls(event):
    """ show subdirs in working directory. """
    event.reply(" ".join(os.listdir(_cfg.workdir)))

ls.perm = "USER"

def man(event):
    """ show descriptions of the available commands. """
    bot = fleet[0]
    for cmnd in bot._cmds:
        description = bot._descriptions.get(cmnd, "")
        if description:
            event.reply("%-20s %s" % (cmnd, description))

man.perm = "USER"

def nick(event):
    """ change bot nick on IRC. """
    bot = fleet[0]
    if event._parsed.args:
        nick = event._parsed.args[0]
        bot.nick(nick)
        bot.cfg.save()
        event.reply("nick changed to %s" % nick)
    else:
        event.reply("nick <name>")
        return

nick.perm = "OPER"

def pid(event):
    """ show pid of the bot. """
    event.reply("pid is %s" % str(os.getpid()))

pid.perm = "USER"

def permissions(event):
    """ show permissions granted to a user. """
    bot = fleet[0]
    perms = []
    for name, funcs in bot._handlers.items():
        if event._parsed.rest in name:
            perms.extend([str(x.perm) for x in funcs if "perm" in dir(x)])
    event.reply("permissions: %s" % ",".join(set(perms)))

permissions.perm = "USER"

def ps(event):
    """ show running threads. """
    launcher = Launcher()
    up = elapsed(int(time.time() - _starttime))
    result = []
    for thr in sorted(launcher.running(), key=lambda x: name(x)):
        obj = Object(vars(thr))
        if "sleep" in obj:
            up = obj.sleep - int(time.time() - obj._lasttime)
        else:
            up = int(time.time() - _starttime)
        thrname = name(thr)[1:] + ")"
        result.append((up, thrname, obj))
    nr = 0
    for up, thrname, obj in sorted(result, key=lambda x: x[0]):
        nr += 1
        res = psformat % (nr, elapsed(up), thrname[:30])
        event.reply(res.rstrip())

ps.perm = "PS"

def meet(event):
    """ create an user record. """
    try:
        origin = event._parsed.rest
    except:
        event.reply("meet <nick> [<perm1> <perm2>]")
        return
    orig = origin.strip()
    if not allowchar(orig):
        event.reply("%s is not allowed as a username" % orig)
        return
    if orig:
        u = users.meet(orig)
        if u:
            event.reply("user %s created" % origin)
        else:
            event.reply("%s is already introduced." % origin)

meet.perm = "MEET"

def mods(event):
    """ show available modules. """
    bot = fleet[0]
    modnames = [x.split(".")[-1] for x in bot.modules("bot")]
    res = []
    for name in modnames:
        n = "bot.%s" % name
        try:
            mod = bot.direct(n)
        except ImportError:
            pass
        res.append("%-20s%s" % (n, mod.__doc__))
    event.reply("\n".join(res))

def perm(event):
    """ add/change permissions of an user. """
    try:
        origin, perms = event._parsed.args
    except:
        event.reply("perm <origin> <perm>")
        return
    u = users.perm(origin, perms)
    if not u:
        event.reply("can't find a user matching %s" % origin)
        return
    event.ok(origin)

perm.perm = "OPER"

def perms(event):
    """ show permission of user. """
    u = users.fetch(event.origin)
    if u:
        event.reply("permissions are %s" % ",".join(u.perms))

perms.perm = "USER"

def real_reboot():
    """ actual reboot. """
    reset()
    os.execl(sys.argv[0], *(sys.argv + ["-r",]))

real_reboot.perm = "OPER"

def reboot(event):
    """ reboot the bot, allowing statefull reboot (keeping connections alive). """
    if not _cfg.reboot:
        event.reply("# reboot is not enabled.")
        return
    event.announce("rebooting")
    real_reboot()

reboot.perm = "OPER"

def reload(event):
    """ reload a plugin. """
    if not event._parsed.rest:
        event.reply(",".join([x.split(".")[-1] for x in bot.modules("bot")]))
        return
    for modname in bot.modules("bot"):
        if event._parsed.rest not in modname:
            continue
        try:
            mod = bot.reload(modname, False)
            if mod:
                event.ok(modname)
            else:
                event.reply("no %s module foudn to reload." % modname)
        except (AttributeError, KeyError) as ex:
            event.reply("%s %s" % (modname, str(ex)))

reload.perm = "RELOAD"

def restore(event):
    """ set deleted=False in selected records. """
    nr = 0
    event.nodel = True
    for obj in db.selected(event):
        obj.deleted = False
        obj.sync()
        nr += 1
    event.ok(nr)

restore.perm = "RESTORE"

def rm(event):
    """ set deleted flag on objects. """
    nr = 0
    for obj in db.selected(event):
        obj.deleted = True
        obj.sync()
        nr += 1
    event.ok(nr)

rm.perm = "RM"

def rss(event):
    """ add a rss url. """
    if not event._parsed.rest:
        event.reply("rss <item>")
        return
    o = Rss()
    o.rss = event._parsed.rest
    o.service = "rss"
    o.save()
    event.ok(1)

rss.perm = "USER"

def silent(event):
    """ put a bot into silent mode. """
    for bot in fleet:
        if event.id == bot.id:
            bot.cfg.silent = True
            bot.cfg.sync()
            event.reply("silent mode enabled.")
            return
    event.reply("no %s bot found in the fleet." % event.id)

silent.perm = "SILENT"

def shop(event):
    """ add a shopitem to the shopping list. """
    if not event._parsed.rest:
        event.reply("shop <item>")
        return
    o = Shop()
    o.shop = event._parsed.rest
    o.save()
    event.ok(1)

shop.perm = "USER"

def show(event):
    """ show dumps of basic objects. """
    mods = get("<module", "bot", g=globals())
    if not event._parsed.rest:
        event.reply(",".join(mods))
        return
    try:
        item, value = event._parsed.rest.split(" ", 1)
    except:
        item = event._parsed.rest
        value = None
    if item == "bot":
        event.reply(fleet[0])
        return
    if item == "fleet":
        for bot in fleet:
            if value:
                val = getattr(bot, value, None)
                event.reply("%s=%s" % (value, val))
            else:
                event.reply(bot)
        return
    obj = getattr(mods, item, None)
    if value:
        val = getattr(obj, value, None)
        event.reply("%s: %s" % (value, val))
    else:
        event.reply(obj)

show.perm = "SHOW"

def start(event):
    """ start a plugin. """
    bot = fleet[0]
    modnames = [x.split(".")[-1] for x in bot.modules("bot")]
    if not event._parsed.rest:
        res = [x.split(".")[-1] for x in modnames]
        event.reply("choose one of %s" % ",".join(sorted(res)))
        return
    modname = event._parsed.args[0]
    name = "bot.%s" % modname
    if name not in modnames:
        event.reply("no %s module found." % name)
        return
    mod = bot.load(name, force=True)
    if "init" in dir(mod):
        mod.init(event)
    event.ok(name)

start.perm = "START"
start._threaded = True

def stop(event):
    """ stop a plugin. """
    if not event._parsed.rest:
        event.reply("stop what ?")
        return
    bot = fleet[0]
    name = "bot.%s" % event._parsed.args[0]
    if name not in bot.modules("bot"):
        event.reply("choose one of %s" % ",".join(sorted(bot.modules("bot"))))
        return
    mod = loader.load_mod(name)
    if "shutdown" in dir(mod):
        mod.shutdown(event)
    event.ok(event.txt)

stop._threaded = True
stop.perm = "STOP"

def synchronize(event):
    """ synchronize rss feeds (fetch but don't show). """
    rss = RSS()
    for client in rss.clients:
        #c = construct(client)
        seen = client.synchronize()
        event.reply("%s urls updated" % len(seen.urls))

synchronize.perm = "SYNCHRONIZE"

def test(event):
    """ echo origin. """
    event.reply("hello %s" % event.origin)

test.perm = "USER"

def tests(event):
    users.uber("test@shell")
    events = []
    bot = HandlerBot()
    bot.load_mod("bot.cmd")
    bot.start()
    for cmnd in bot._cmds:
        if cmnd in ["all", "tests", "tinder"]:
            continue
        e = Event()
        e.id = bot.id
        e.origin = "test@shell"
        e.txt = cmnd + " " + randomname()
        e.txt = e.txt.strip()
        bot.put(e)
        events.append(e)
    for e in events:
        e.wait()
    bot.stop()
    event.reply("#%s commands run." %  len(bot._cmds))
    
tests.perm = "OPER"

def timer(event):
    """ timer command to schedule a text to be printed on a given time. stopwatch to measure elapsed time. """
    if not event._parsed.rest:
        event.reply("timer <string with time> <txt>")
        return
    seconds = 0
    line = ""
    try:
        target = parse_time(event.txt)
    except ValueError:
        event.reply("can't detect time")
        return
    if not target or time.time() > target:
        event.reply("already passed given time.")
        return
    t = Timed()
    t.services = "clock"
    t.prefix = "timer"
    t.txt = event._parsed.rest
    t.time = target
    t.done = False
    t.save()
    event.ok(time.ctime(target))

timer.perm = "USER"

def tinder(event):
    """ loop an command nr of times. """
    from .ldr import loader
    if not event._parsed.args:
        event.reply('tinder <cmnd> <nr>')
        return
    name = event._parsed.args[0]
    try:
        nr = int(event._parsed.args[1])
    except (IndexError, ValueError):
        nr = 10
    mod = loader.direct("bot.cmd")
    if mod:
        cmd = getattr(mod, name, None)
        if not cmd:
            event.reply("can't find %s" % cmd)
            return
    e = Event()
    e.origin = "test@shell"
    for x in range(nr):
        launcher.launch(cmd, event)

def today(event):
    """ show last week's logged objects. """
    event._parsed.start = to_day(day())
    event._parsed.end = time.time()
    nr = 0
    for obj in db.selected(event):
        nr += 1
        event.display(obj, str(nr))

today.perm = "USER"

def todo(event):
    """ log a todo item. """
    if not event._parsed.rest:
        event.reply("todo <item>")
        return
    o = Todo()
    o.todo = event._parsed.rest
    o.save()
    event.ok(1)

todo.perm = "USER"

def tomorrow(event):
    """ show todo items for tomorrow. """
    if not event._parsed.rest:
        event.reply("tomorrow <item>")
        return
    o = Tomorrow()
    o.tomorrow = event.txt.replace("tomorrow", "").strip()
    o.save()
    event.ok(1)

tomorrow.perm = "USER"

def u(event):
    """ show user selected by userhost. """
    if not event._parsed.rest:
        event.reply("u <origin>")
        return
    origin = event._parsed.args[0]
    u = users.fetch(origin)
    if u:
        event.reply(u.pure())
    else:
        event.reply("no user %s found." % origin)

u.perm = "USER"

def w(event):
    """ show user data. """
    u = users.fetch(event.origin)
    if u:
        event.reply(u)
    else:
        event.reply("no matching user found.")

w.perm = "USER"

def uptime(event):
    """ show uptime. """
    event.reply("uptime is %s" % elapsed(time.time() - _starttime))

uptime.perm = "USER"

def version(event):
    """ show version. """
    event.reply("%s #%s - %s" % (_cfg.name, __version__, __txt__))

version.perm = "USER"

def watch(event):
    """ add a file to watch (monitor and relay to channel). """
    if not event._parsed.rest:
        event.reply("watch <item>")
        return
    o = Watch()
    o.watch = event._parsed.rest
    o.save()
    event.ok(1)

watch.perm = "WATCH"

def week(event):
    """ show last week's logged objects. """
    event._parsed.start = to_day(day()) - 7 * 24 * 60 * 60
    event._parsed.end = time.time()
    nr = 0
    for obj in db.selected(event):
        nr += 1
        event.display(obj, str(nr))

week.perm = "USER"

def whoami(event):
    """ show origin. """
    event.reply("you have origin %s" % event.origin)

whoami.perm = "USER"

def yesterday(event):
    """ show last week's logged objects. """
    event._parsed.start = to_day(day()) - 24 * 60 * 60
    event._parsed.end = to_day(day())
    nr = 0
    for obj in db.selected(event):
        nr += 1
        event.display(obj, str(nr))

yesterday.perm = "USER"
