#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BOTJE - Framework to program bots !!
#
# setup.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 22-12-2017 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" setup.py """

import os
import sys

if sys.version_info.major < 3:
    print("you need to run BOTJE with python3")
    os._exit(1)

try:
    use_setuptools()
except:
    pass

try:
    from setuptools import setup
except Exception as ex:
    print(str(ex))
    os._exit(1)

setup(
    name='botje',
    version='51',
    url='https://bitbucket.org/bthate/botje',
    author='Bart Thate',
    author_email='bthate@dds.nl',
    description="Framework to program bots !!",
    license='MIT',
    include_package_data=False,
    zip_safe=False,
    install_requires=[],
    scripts=["bin/botje", "bin/botje-irc", "bin/botje-toudp", "bin/botje-xmpp"],
    packages=['bot'],
    long_description='''
BOTJE
#####

you might need to do one of the following if the bot doesn't work:

::

 bart@okdan:~/botje$ export PYTHONPATH="."
 bart@okdan:~/botje$ export PYTHONIOENCODING="utf-8"

intro
=====

| BOTJE is a python3 framework which you can use to program bots.
| BOTJE makes it possible to program your own module enabling your own commands.
| BOTJE contains serveral modules that can be used to program.
| BOTJE has a MIT no-clause :ref:`license <license>`.

install
=======

clone the source:

:: 

 bart@okdan:~$ hg clone https://bitbucket.org/bthate/botje


another option is to download with pip3 and install globally:

::

 bart@okdan:~$ pip3 install botje --upgrade


irc
===

use -n <nick>, -s <server>, -c <channel> options to make the bot join the network:

::

 bart@okdan:~$ bot-irc -n botje -s irc.freenode.net -c \#botje

users
=====

One needs to add a users origin to be able to give the bot commands. One can add a user with the meet command:

::

 bart@okdan:~$ bot meet user@server
 user user@server created

To give the user a permission you can use the perm command:

::

 bart@okdan:~$ bot perm user@server ps
 ok user@server

The default shell user is root@shell and gives access to all the commands that are available.

programming
===========

| User programmed modules are read from the ~/.mods directory, you can put your .py files over there.

example:

::

     def hi(event):
         event.reply("hi %s" % event.origin)

modules
=======

.. csv-table:: List of Modules
   :header: "Name", "Description"
   :widths: auto

   bot.clk,              timer/repeater/timed daemon (clock). 
   bot.cmd,              basic commands (commands).
   bot.dbz,              access saved json files (databases).
   bot.dcc,              direct client to client bot (DCC). 
   bot.eml,              botlib basic commands (email).
   bot.evt,              event class (event).
   bot.hdl,              event handler based on queue.Queue (handler). 
   bot.irc,              irc bot (IRC).
   bot.lcs,              license command (license).
   bot.ldr,              load modules (loader).
   bot.log,              log module to set standard format of logging (log).
   bot.obj,              object class with save/load of json files (object).
   bot.raw,              raw output using print (raw).
   bot.rss,              rss module (needs feedparser). 
   bot.shl,              shell related functionality (shell). 
   bot.slt,              select based loop (select). 
   bot.thr,              threading module (launcher).
   bot.tmd,              timed helper functions (timed).
   bot.trc,              stack trace helper functions (trace).
   bot.udp,              relay txt through a udp port listener (UDP). 
   bot.url,              functions that fetch data from url (url).
   bot.xmpp,             xmpp bot (requires sleekxmpp). 

commands
========

Over 60 commands are now available in BOTJE.

.. csv-table:: List of Commands
   :header: "Name", "Description"
   :widths: auto

   announce,             announce text on all channels in fleet.
   begin,                begin stopwatch.
   cfg,                  edit config files.
   cmds,                 show list of commands.
   deleted,              show deleted records.
   delperm,              delete permissions of an user.
   edit,                 edit and save objects.
   end,                  stop stopwatch.
   exit,                 stop the bot.
   find,                 present a list of objects based on prompt input.
   first,                show the first record matching the given criteria.
   fix,                  fix a object by loading and saving it.
   last,                 show last objectect matching the criteria.
   load,                 force a plugin reload.
   log,                  log some text.
   loud,                 disable silent mode of a bot.
   ls,                   show subdirs in working directory.
   man,                  show descriptions of the available commands.
   meet,                 create an user record.
   mods,                 show available modules.
   nick,                 change bot nick on IRC.
   perm,                 add/change permissions of an user.
   permissions,          show permissions granted to a user.
   perms,                show permission of user.
   pid,                  show pid of the bot.
   ps,                   show running threads.
   reboot,               allowing statefull reboot (keeping connections alive).
   reload,               reload a plugin.
   restore,              set deleted=False in selected records.
   rm,                   set deleted flag on objects.
   rss,                  add a rss url.
   shop,                 add a shopitem to the shopping list.
   show,                 show dumps of basic objects.
   silent,               put a bot into silent mode.
   start,                start a plugin.
   stop,	         stop a plugin.
   synchronize,          synchronize rss feeds (fetch but don't show).
   test,                 echo origin.
   timer,                timer command to schedule a text to be printed on a given time. stopwatch to measure elapsed time.
   tinder,               loop an command nr of times.
   today,                show last week's logged objects.
   todo,                 log a todo item.
   tomorrow,             show todo items for tomorrow.
   u,                    show user selected by userhost.
   uptime,               show uptime.
   version,              show version.
   w,                    show user data.
   watch,                add a file to watch (monitor and relay to channel).
   week,                 show last week's logged objects. 
   whoami,               show origin.
   yesterday,            show last week's logged objects.
   mbox,                 convert emails to botlib objects.

contact
=======

You can email me or contact me on IRC/freenode/#dunkbots.

| Bart Thate
| botfather on #dunkbots irc.freenode.net
| bthate@dds.nl, thatebart@gmail.com
| hg clone https://bitbucket.org/bthate/botje


''',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Topic :: Utilities'],
)
